import typer

from .nautikos import Nautikos

app = typer.Typer()

nautikos = Nautikos()


@app.command()
def main(
    repository: str,
    tag: str,
    env: str = typer.Option(None),
    labels: str = typer.Option(None),
    config: str = typer.Option("nautikos.yaml"),
    dry_run: bool = typer.Option(False, "--dry-run"),
):
    nautikos.set_dry_run(dry_run)
    nautikos.load_config(config)
    nautikos.update_manifests(
        repository, tag, environment=env, labels=labels.split(",")
    )



from __future__ import annotations

import os
import pathlib
import sys
from typing import TypedDict

from .manifests import get_manifest
from .yaml import yaml


class ManifestConfig(TypedDict):
    path: str
    type: str
    labels: list[str]
    repositories: list[str]


class EnvironmentConfig(TypedDict):
    name: str
    manifests: list[ManifestConfig]


class ConfigData(TypedDict):
    environments: list[EnvironmentConfig]


class Nautikos:
    def __init__(self) -> None:
        self._workdir: pathlib.Path = pathlib.Path(".")
        self._dry_run: bool = False
        self._environments: list[EnvironmentConfig] = []

    def set_dry_run(self, dry_run: bool) -> None:
        self._dry_run = dry_run

    def load_config(self, path: str) -> None:
        self._workdir = pathlib.Path(path).parent
        with open(path, "r") as f:
            config_data: ConfigData = yaml.load(f)
        self._environments = config_data["environments"]

    def update_manifests(
        self,
        repository: str,
        new_tag: str,
        environment: str | None = None,
        labels: list[str] | None = None,
    ) -> None:
        """Updates image tags of given repository to new tag

        If environment is passed, only environments with matching name are modified;
        otherwise all environments are modified.

        If label is passed, only manifests with matching label are modified; otherwise
        all manifests in selected environments are modified.
        """
        # Create list of environments
        envs: list[EnvironmentConfig] = []
        for env in self._environments:
            if not environment or env["name"] == environment:
                envs.append(env)
        if len(envs) == 0:
            raise Exception(f"Oops! No environments with name '{environment}' found...")

        # Create list of manifests
        manifests: list[ManifestConfig] = []
        for env in envs:
            for manifest in env["manifests"]:
                if not labels or set(labels).issubset(set(manifest["labels"])):
                    manifests.append(manifest)
        if len(manifest) == 0:
            raise Exception(
                f"Oops!! No manifest found; environment={environment}"
                f"labels={labels}"
            )

        # Modify manifests
        for manifest_config in manifests:
            self._modify_manifest(
                manifest_config["type"],
                manifest_config["path"],
                repository,
                new_tag,
            )

    def _modify_manifest(self, type: str, path: str, repository: str, tag: str) -> None:
        manifest = get_manifest(type)
        with open(os.path.join(self._workdir, pathlib.Path(path)), "r") as s:
            manifest.load(s)
        manifest.modify(repository, tag)
        if self._dry_run:
            manifest.write(sys.stdout)
        else:
            with open(os.path.join(self._workdir, pathlib.Path(path)), "w") as s:
                manifest.write(s)
                print(self._log_string(repository, tag, pathlib.Path(path)))

    @staticmethod
    def _log_string(repository: str, tag: str, path: str | pathlib.Path) -> str:
        return f"Modified tag for {repository} to {tag} in {path}"



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['nautikos']

package_data = \
{'': ['*']}

install_requires = \
['ruamel-yaml>=0.17.21,<0.18.0', 'typer>=0.7.0,<0.8.0']

entry_points = \
{'console_scripts': ['nautikos = nautikos.cli:app']}

setup_kwargs = {
    'name': 'nautikos',
    'version': '0.2.0',
    'description': 'A CD tool for updating image tags in Kubernetes manifests',
    'long_description': "# Nautikos \n\n*A lightweight CI/CD tool for updating image tags in Kubernetes manifests.* \n\n[![PyPI version](https://badge.fury.io/py/nautikos.svg)](https://badge.fury.io/py/nautikos)\n[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)\n[![Python versions](https://img.shields.io/pypi/pyversions/nautikos)]()\n\n## Rationale\n\nA GitOps CI/CD process usually uses a deployment or ops repo containing Kubernetes manifests for multiple services and environments. Tools like *Argo-CD* or *Flux* then track these repo's, and apply any changes to the cluster. When a new image of an application is created, you want the corresponding tags to be updated in the manifests. Doing this manually is error prone. Having to write logic in every repo or pipeline to perform this is tedious. \n\nThis is where Nautikos comes in. \n\n## Installation \n\n```bash\npip install nautikos\n```\n\n## Basic usage \n\nNautikos is configured through a YAML-file (`./nautikos.yaml`), that specifies where the manifests for the different images and environments can be found: \n\n```yaml\nenvironments: \n- name: prod \n  manifests: \n  - path: prod/app1/deployment.yaml  # Path relative to configuration file\n    type: kubernetes  # Type can be 'kubernetes' or 'kustomize'\n    labels: \n    - app1\n    - refs/head/main\n  - path: prod/app2/kustomize.yaml\n    type: kustomize\n    labels:  # Optional specification of labels for more granular control\n    - app2\n    - refs/head/master\n- name: dev\n  manifests: \n  - path: dev/app1/deployment.yaml\n    type: kubernetes\n    labels: \n    - app1\n    - refs/head/dev\n  - path: dev/app3/feature-A/deployment.yaml\n    type: kubernetes\n    labels: \n    - app1\n    - refs/head/feature-A\n```\n\nNext, you can run Nautikos to update the image tags of specific images in different environments.\n\n```bash\nnautikos my-repo 1.2.3  # Updates all occurences of `my-repo` to `1.2.3`\nnautikos --env prod my-other-repo 2.3.4  # Updates all occurences of `my-other-repo` to `2.3.4` in `prod/app1/deployment.yaml` and `prod/app2/deployment`\nnautikos --env dev --label app1 my-other-repo 4.5.6  # Updates all occurences of `my-other-repo` to `4.5.6` in `dev/app1/deployment.yaml`\nnautikos --label 'app1,refs/head/main' my-other-repo 5.6.7  # Updates all occurences of `my-other-repo` to `5.6.7` in `prod/app1/deployment.yaml`\n```\n\n## Supported tools\n\nThe tool works with standard **Kubernetes** manifests and **Kustomize** - **Helm** might be added later. Each have their own format for defining image tags. \n\n```yaml\n# Kubernetes manifests\nspec:\n  template:\n    spec:\n      containers:\n      - image: some-repository:tag\n\n# Kustomize\nimages: \n- name: some-repository\n  newTag: tag \n```\n\n## Advanced usage\n\nNautikos takes several options: \n\n* `--dry-run`: prints the modified files to stdout, but doesn't edit in place \n* `--config config-file.yaml`: path to config YAML, default is `./nautikos.yaml`\n\n## Alternatives \n\nThere are basically three alternatives to do the same thing: \n\n* **Update manifests manually** - of course this works, but this is not really proper CD\n* **Write your own bash scripts in a pipeline using a tool like `sed`** - This works, but having to write this logic for every project is tedious. \n* **Use a tool like [Argo-CD Image updater](https://argocd-image-updater.readthedocs.io/en/stable/)** - very nice, but a bit heavy-weight, not very actively developed, and doesn't seem to support Azure Container Registry. \n\n## Notes \n\nMultiple YAML docs in one file is not yet supported. \n\n## Dependencies \n\n* **`typer`** - for creating a CLI \n* **`ruamel.yaml`** - for handling YAML files while maintaining ordering and comments\n",
    'author': 'Jan Hein de Jong',
    'author_email': 'janhein.dejong@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)



import abc
from typing import Any, TextIO, TypedDict

from .yaml import yaml


class Image(TypedDict):
    repository: str
    tag: str | None


class AbstractManifest(abc.ABC):
    def load(self, stream: TextIO) -> None:
        self._data = yaml.load(stream)

    def write(self, stream: TextIO) -> None:
        yaml.dump(self.data, stream)

    @property
    def data(self) -> Any:
        if self._data:
            return self._data
        else:
            raise Exception("You must first load a manifest")

    @abc.abstractmethod
    def modify(self, repository: str, new_tag: str) -> None:
        ...

    @abc.abstractmethod
    def get_images(self) -> list[Image]:
        ...


KubernetesImageDefinition = str


class KubernetesContainer(TypedDict):
    image: KubernetesImageDefinition


class KubernetesManifest(AbstractManifest):
    def modify(self, repository: str, new_tag: str) -> None:
        for container in self._get_containers():
            if repository == self._parse_image(container["image"])["repository"]:
                container["image"] = self._unparse_image(
                    {"repository": repository, "tag": new_tag}
                )

    def get_images(self) -> list[Image]:
        return [
            self._parse_image(container["image"])
            for container in self._get_containers()
        ]

    def _get_containers(self) -> list[KubernetesContainer]:
        return self.data["spec"]["template"]["spec"]["containers"]

    def _parse_image(self, image: KubernetesImageDefinition) -> Image:
        if ":" in image:
            repository, tag = tuple(image.split(":"))
        else:
            repository, tag = image, None
        return {"repository": repository, "tag": tag}

    def _unparse_image(self, image: Image) -> str:
        return f"{image['repository']}:{image['tag']}"


class KustomizeImageDefinition(TypedDict):
    name: str
    newTag: str


class KustomizeManifest(AbstractManifest):
    def modify(self, repository: str, new_tag: str) -> None:
        for kustomize_image in self._data["images"]:
            if repository == kustomize_image["name"]:
                kustomize_image["newTag"] = new_tag

    def get_images(self) -> list[Image]:
        return [self._parse_image(image) for image in self.data["images"]]

    def _parse_image(self, image: KustomizeImageDefinition) -> Image:
        return {"repository": image["name"], "tag": image["newTag"]}


def get_manifest(type: str) -> AbstractManifest:
    if type == "kubernetes":
        return KubernetesManifest()
    elif type == "kustomize":
        return KustomizeManifest()
    elif type == "helm":
        raise Exception("Helm manifests are not yet implemented.")
    else:
        raise Exception(f"'{type}' is not a correct manifest type.")
